; Part of mabloss - MAnipulating BLOggs Simply in Scheme 
; Copyright 2004 MJ Ray mjr at towers.org.uk
; Licensed under the GNU GPL, see http://www.gnu.org/

(module mabloss mzscheme

(require (lib "list.ss") (lib "date.ss") (lib "pregexp.ss") (lib "file.ss") (lib "xml.ss" "xml"))

(provide read-prompt conf-ref seq unpick body find-in
   copyright datestamp datesort-2822s atom-2822 clean-2822
   read-2822 read-2822s write-2822 index-2822 field-2822 id-2822
   file->xml xml->file
   div->title div->2822 2822->div
   divs->xhtml xhtml->divs
   rssitem->title rssitem->2822 2822->rssitem
   rssitems->rss rss->rssitems
   file->2822 2822->file datesort-files dir->files)

(define copyright " - v1.1 - Copyright 2004 MJ Ray\nLicensed under the GNU GPL, see http://www.gnu.org/licenses/gpl.html\n\n")

(define digits (string->list "0123456789"))

(define (number->chars num len)
  (if (= len 0)
    '()
    (let* ((c (- len 1)) (e (expt 10 c)))
      (cons
        (list-ref digits (quotient num e))
        (number->chars (modulo num e) c)))))

(define (valid-date? str)
  (pregexp-match "^\\d+-\\d\\d-\\d\\dT\\d\\d:\\d\\d:\\d\\d" str))

(define (datestamp when)
  (let ((d (seconds->date when)))
    (list->string
      (cdr
        (apply append
          (map
            (lambda (pre fn wdth) (cons pre (number->chars (fn d) wdth)))
            '(err #\- #\- #\T #\: #\:)
            (list date-year date-month date-day date-hour date-minute date-second)
            '(4 2 2 2 2 2)))))))

;-----
; List manipulators

(define (seq s e)
    (if (= s e) (list s) (cons s (seq (+ s 1) e))))

(define (unpick l i)
    (if (= i 1) (cdr l) (cons (car l) (unpick (cdr l) (- i 1)))))

(define (body l)
  (apply string-append (filter string? l)))

(define (find-in hay key)
    (cond
      ((or (not (list? hay)) (null? hay)) #f)
      ((and (list? (car hay)) (not (null? (car hay))) (equal? key (caar hay)))
        hay)
      (#t (find-in (cdr hay) key))))

;-----
; User input

(define (read-prompt prompt all? . port)
  (display prompt)
  (if all?
    (apply read-string (cons 65535 port))
    (apply read-line port)))

;-----
; Config file handling

(define conf-ref
  (let ((conf (call-with-input-file (string-append (getenv "HOME") "/.mablossrc") read)))
    (lambda (symbol)
      (cadr (or (assoc symbol conf) '("" ""))))))

(date-display-format (conf-ref 'dates))

;-----
; RFC-2822 representation handling
;
; The representation is a list containing an alist of
; the headers and upto 64k of the body string.

(define spaces '(#\space #\tab))

(define (atom-2822 str)
  (pregexp-replace* "[\\s[:cntrl:]\\(\\)<>\\[\\]:;@\\,]+" str ""))

(define (read-2822 . port) ; . port -> alist-strings-list
  (let ((p (if (null? port) (current-input-port) (car port))))
    (list
      (let rdr ((char (read-char p)) (head '()) (line '()) (field '()))
        (cond
          ((eof-object? char) head)
          ((equal? #\newline char)
            (cond
              ((null? field) (reverse head))
              ((member (peek-char p) spaces)
                (rdr (read-char p) head line field))
              (#t
                (rdr
                  (read-char p)
                  (cons (list line (list->string (reverse field))) head)
                  '()
                  '()))))
          ((and (equal? #\: char) (null? line))
            (read-char p)
            (rdr (read-char p) head (list->string (reverse field)) '()))
          (#t (rdr (read-char p) head line (cons char field)))))
      (let rdr ((char (read-char p)) (buf '()))
        (cond
          ((or (eof-object? char) (and (equal? #\. char) (equal? #\newline (car buf)) (equal? #\newline (peek-char p))))
            (read-char p)
            (list->string (reverse buf)))
          ((and (equal? #\. char) (equal? #\newline (car buf)) (equal? #\. (peek-char p)))
            (display ".")
            (rdr (read-char p) buf))
          (#t (rdr (read-char p) (cons char buf))))))))

(define (read-2822s . port)
  (let rdr ((p (if (null? port) (current-input-port) (car port)))
            (buf '()))
    (if (eof-object? (peek-char p))
      buf
      (rdr p (cons (read-2822 p) buf)))))

(define (write-2822 msg . port) ; alist-strings-list . port -> void
  (let ((p (if (null? port) (current-output-port) (car port))))
    (for-each
      (lambda (i) (display (string-append (car i) ": " (cadr i) "\n") p))
      (car msg))
    (newline p)
    (for-each (lambda (i) (display i p)) (cdr msg))
    (display "\n.\n" p)))

(define (index-2822 key msgs)
  (map
    (lambda (i)
      (cadr
        (or
          (and (list? i) (not (null? i)) (assoc key msgs)
          (list #f (number->string (current-seconds)))))))
    msgs))

(define (field-2822 field msg default)
  (cadr (or (assoc field (car msg)) (list #f default))))

(define (id-2822 msg)
  (let ((rid (field-2822 "Message-ID" msg (field-2822 "Message-Id" msg (format "<~a@~a>" (current-seconds) (conf-ref 'id-host))))))
    (substring rid 1 (- (string-length rid) 1))))

(define (2822-addhdr msg hdr val)
  (cons (cons (list hdr val) (car msg))
        (cdr msg)))

(define (clean-2822 msg)
  (let ((imsg (if (field-2822 "Message-ID" msg #f)
                  msg
                  (2822-addhdr msg "Message-ID" (id-2822 msg)))))
    (if (field-2822 "Date" imsg #f)
        imsg
        (2822-addhdr msg "Date" (datestamp (current-seconds))))))

(define (laterdate? a b)
  (string>? (field-2822 "Date" a "") (field-2822 "Date" b "")))

(define (datesort-2822s msgs)
  (quicksort msgs laterdate?))

;-----
; xml-file conversions

(define (file->xml file)
  (filter (lambda (i) (or (list? i) (symbol? i))) (xml->xexpr (document-element (call-with-input-file file read-xml)))))

(define (xml->file sxml fname)
  (write-xml/content
     (xexpr->xml sxml)
     (open-output-file fname 'replace)))

;-----
; 2822-div conversion functions

(define (div->title item)
  (regexp-replace "\n" (body (car (find-in item 'h4))) " "))

(define (div->from item)
  (regexp-replace "\n" (body (car (or (find-in item 'address) (list (list (getenv "EMAIL")))))) " "))

(define (is-class? class)
  (lambda (ele)
    (and
      (list? ele)
      (> (length ele) 1)
      (member (list 'class class) (cadr ele)))))

(define (div->date item)
  (let ((d (filter (is-class? "date") item)))
    (if (or (null? d) (not (valid-date? (body (car d)))))
      (datestamp (current-seconds))
      (body (car d)))))

(define (div->id item)
  (cadr (or (assoc 'id (cadr item)) (list #f (current-seconds)))))

;FIXME: Far too fragile
(define (div->url item fname id)
  (let ((d (filter (is-class? "permalink") item)))
    (if (null? d)
      (string-append (conf-ref 'xhtml-base) fname "#" id)
      (cadar (cadr (caddr (car d)))))))

(define (div->2822 item fname)
  (let ((id (div->id item)))
    `((("From" ,(div->from item))
       ("Subject" ,(div->title item))
       ("Date" ,(div->date item))
       ("Message-Id" ,(string-append "<" id ">"))
       ("X-URL" ,(div->url item fname id)))
      ,(pregexp-replace "<div>([[:alpha:][:^alpha:]]*)</div>" (xexpr->string (car (find-in item 'div))) "\\1"))))

(define (2822->div data urlbase)
  (let* ((id (id-2822 data))
         (title (field-2822 "Subject" data "Title unknown"))
         (from (field-2822 "From" data (or (getenv "EMAIL") "Unknown")))
         (desc (cadr data))
         (date (field-2822 "Date" data ""))
         (link (field-2822 "X-URL" data (string-append urlbase "#" id))))
    (if (getenv "DEBUG") (display (format "2822->div(~a)\n" id) (current-error-port)))
  `(div ((id ,id))
      (h4 ,title)
      (address ,from)
      ,(if (and (> (string-length desc) 1) (string=? (substring desc 0 1) "<"))
           (with-handlers ((not-break-exn? (lambda (p) `(div (p ,desc))))) (read-xml/element (open-input-string (string-append "<div>" desc "</div>"))))
           `(div (p ,desc)))
      (div ((class "permalink")) (a ((href ,link)) "More..."))
      (div ((class "date")) ,(if (valid-date? date) date (datestamp (current-seconds)))))))

;-----
; div-xhtml conversion functions

(define (divs->xhtml items i)
  (cond
    ((or (not (list? i)) (< (length i) 2)) i)
    ((equal? '((id "blog")) (cadr i)) `(div ((id "blog")) ,@items))
    ((list? i) (map (lambda (j) (divs->xhtml items j)) i))
    (#t i)))

(define (is-blog? ele)
  (and (list? ele) (> (length ele) 1) (equal? '((id "blog")) (cadr ele))))

(define (xhtml->divs page)
  (filter list? (cddar (filter is-blog? (car (find-in page 'body))))))

;-----
; 2822-rssitem conversion functions

(define (rssitem->title item)
  (regexp-replace "\n" (body (car (or (find-in item 'title) '(("Untitled"))))) " "))

#cs(define (rssitem->2822 item rsss)
  (if (getenv "DEBUG") (display "rssitem->2822 " (current-error-port)))
  (let* ((chan (car (or (find-in rsss 'channel) '(()))))
         (id (string-append "<" (atom-2822 (body (car (or (find-in item 'link) '(("NULL")))))) "@" (atom-2822 (body (car (or (find-in chan 'link) (find-in chan 'title) '(("NULL")))))) ">")))
    (if (getenv "DEBUG") (display item (current-error-port)))
    `((("From" ,(body (car (or (find-in item 'dc:Author) (find-in chan 'dc:Author) (find-in chan 'author) (find-in chan 'title) (list (list (getenv "EMAIL")))))))
       ("Subject" ,(rssitem->title item))
       ("Date" ,(body (car (or (find-in item 'dc:Date) (find-in item 'pubDate) (list (list (datestamp (current-seconds))))))))
       ("Message-ID" ,id)
       ("X-URL" ,(body (car (or (find-in item 'link) '(("")))))))
      ,(body (car (or (find-in item 'content:encoded) (find-in item 'description) '((""))))))))

#cs(define (2822->rssitem data)
  (if (getenv "DEBUG") (display "2822->rssitem: " (current-error-port)))
  (let ((id (id-2822 data))
        (title (field-2822 "Subject" data "Title unknown"))
        (from (field-2822 "From" data (or (getenv "EMAIL") "Unknown")))
        (desc (cadr data))
        (date (field-2822 "Date" data (date->string (seconds->date (current-seconds)) #t)))
        (link (field-2822 "X-URL" data "")))
    (if (getenv "DEBUG") (display data (current-error-port)))
    `(item ((rdf:about ,link)) (title ,title) (link ,link) (content:encoded ,desc) (dc:Date ,date) (dc:Author ,from))))

;-----
; rssitem-rss conversion functions

#cs(define (index-rssitems items)
  `(items
    (rdf:Seq
      ,@(map
          (lambda (i)
            `(rdf:li ((rdf:resource ,(body (car (find-in i 'link)))))))
          items))))

;FIXME: strips stuff other than the first, channel and item tags
(define (rssitems->rss items rsss)
  (let ((channel (filter list? (car (find-in rsss 'channel)))))
             (append
               (list
                 (car rsss)
                 (cadr rsss)
                 `(channel
                   ,@(filter (lambda (i) (or (null? i) (not (equal? (car i) 'items)))) channel)
                   ,(index-rssitems items)))
               items)))

(define (rss->rssitems rsss)
  (filter list? (or (find-in rsss 'item) (find-in (caddr rsss) 'item))))

;-----
; 2822-file conversions

(define (file->title) "NOT LOADED")

(define (file->2822 fname)
  (call-with-input-file fname read-2822))

(define (2822->file msg dir)
  (let* ((cmsg (clean-2822 msg))
         (id (pregexp-replace* "/" (id-2822 cmsg) ""))
         (fname (string-append dir id ".txt")))
    (if (file-exists? fname)
      (display "Given Message-ID already exists\n")
      (call-with-output-file
        fname
        (lambda (p) (write-2822 cmsg p))))))

;-----
; file-dir conversion

(define (get-times files)
  (map
    (lambda (i)
      (cons (file-or-directory-modify-seconds i) i))
    files))

(define (later? a b)
  (> (car a) (car b)))

(define (datesort-files files)
  (map cdr (quicksort (get-times files) later?)))

(define (dir->files dir)
  (find-files
    file-exists?
    dir))

)