#!/usr/bin/mzscheme -qr
;
; mablossom.scm - Blosxom-like CGI blog generator
; Part of mabloss - MAnipulating BLOggs Simply in Scheme 
; Copyright (C) 2004 MJ Ray mjr at towers.org.uk
; Licensed under the GNU GPL, see http://www.gnu.org/

(require (lib "file.ss") (lib "list.ss") (lib "pregexp.ss") (lib "xml.ss" "xml"))

;-----
; Config file handling

(define conf-ref
  (let ((conf (call-with-input-file "mablossom.conf" read)))
    (lambda (symbol)
      (cadr (or (assoc symbol conf) '("" ""))))))

;-----
; Finding text files

(define (directory-list-recurse dir)
  (find-files
    (lambda (n)
      (and (file-exists? n) (equal? (filename-extension n) "txt")))
    dir))

;-----
; Handling file times

(define (get-times files)
  (map
    (lambda (i)
      (cons (file-or-directory-modify-seconds i) i))
    files))

(define (later? a b)
  (lambda (a b) (> (car a) (car b))))

(define (datesort timedfiles)
  (quicksort timedfiles later?))

;-----
; Reading and writing files

(define (readitem timedfile)
  (call-with-input-file (cdr timedfile)
    (lambda (p)
      (list
        (list "modified" (car timedfile))
        (list "id" (cdr timedfile))
        (list "title" (read-line p))
        (list "link" (read-line p))
        (list "description" (read-string 65535 p))))))

(define (fasttemplate templ filler) ; -> string
  ; fill an output template file
  (let ((ttxt (call-with-input-file templ ; read it in
                (lambda (p) (read-string 65535 p))))
          (safe-filler
            (map (lambda (x)
              (list (car x)
                (if (null? (cdr x)) "" (format "~A" (cadr x)))))
              filler)))
    (let txfn ((tmpl ttxt) (matchers safe-filler))
      (if (null? matchers)
          tmpl
          (txfn (pregexp-replace*
                 (string-append "<\\?mabloss " (caar matchers) "\\?>")
                 tmpl (cadar matchers))
                (cdr matchers))))))

;-----
; Output formatters

(define (item->dldiv item)
  (fasttemplate
    (conf-ref 'itemtemplate)
    item))  

(define (htmlformat items)
  (display
    (fasttemplate
      (conf-ref 'maintemplate)
      (list (list "blog"
        (apply string-append (map item->dldiv items)))))))

(define (rssformat items)
  (display "<?xml version=\"1.0\"?>\n")
  (display-xml/content
   #cs(xexpr->xml
    `(rdf:RDF ((xmlns:rdf "http://www.w3.org/1999/02/22-rdf-syntax-ns#")
               (xmlns "http://purl.org/rss/1.0/"))
       (channel ((rdf:about ,(string-append (conf-ref 'url) (getenv "PATH_INFO")))
         (title ,(conf-ref 'title))
         (link ,(conf-ref 'url))
         (description ,(conf-ref 'description))
         (items
           (rdf:Seq
             ,@(map
                 (lambda (i)
                   `(li ((rdf:about ,(cadr (assoc "link" i))))))
                 items))))
       ,@(map
           (lambda (i)
             `(item ((rdf:about ,(cadr (assoc "link" i))))
               ,(assoc "title" i)
               ,(assoc "link" i)
               ,(assoc "description" i)))
           items)))))

;-----
; Main form

(let ((path (or (getenv "PATH_INFO") (conf-ref 'defaultdir))))
  ((if (equal? "rss" (filename-extension path))
       rssformat
       htmlformat)
    (map
      readitem
      (datesort (get-times (directory-list-recurse (string-append (conf-ref 'home) (path-only path))))))))
